/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.Value;

/**
 * Exploration result for a single solution, i.e. its fitness w.r.t to the investigated optimization
 * dimensions.
 * 
 * @author barner
 */
public class SolutionQuantification {

	/**
	 * Map to store the fitness values for a single solution w.r.t to the investigated optimization
	 * dimensions.
	 */
	private Map<ExplorationTarget<?>, Value<?>> solutionQuantificationMap = new HashMap<>();

	/** Adds a fitness value for a given optimization dimension */
	public <T> void addSingleExplorationResult(final ExplorationTarget<T> target, Value<?> value) {
		solutionQuantificationMap.put(target, value);
	}

	/**
	 * Returns the quantification (the evaluation result as a double value) for the given
	 * {@link ExplorationTarget}.
	 */
	public Double getQuantificationOf(ExplorationTarget<?> expTarget) {
		Value<?> value = getTarget(expTarget);
		if(value != null) {
			return value.getDouble();
		}
		return null;
	}

	/**
	 * Returns the {@link Value} (evaluation result) from the internal map. First it checks whether
	 * the given type is registered in the map. If none is found all entries are checked whether
	 * they are a super type of the given type and the first match is returned.
	 */
	public Value<?> getTarget(ExplorationTarget<?> expTarget) {
		Value<?> value = solutionQuantificationMap.get(expTarget);

		if(value == null) {
			for(ExplorationTarget<?> currExpTgt : solutionQuantificationMap.keySet()) {
				if(expTarget.getClass().isAssignableFrom(currExpTgt.getClass())) {
					value = solutionQuantificationMap.get(currExpTgt);
				}
			}
		}

		return value;
	}

	/** Returns the {@link ExplorationTarget} -> {@link Value} entries of the internal map. */
	public Set<Entry<ExplorationTarget<?>, Value<?>>> entrySet() {
		return solutionQuantificationMap.entrySet();
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String rval = "";
		for(ExplorationTarget<?> target : solutionQuantificationMap.keySet()) {
			rval +=
					(target.getName() + "(id: " + target.getId() + ")" + "=" +
							solutionQuantificationMap.get(target) + ", ");
		}

		return rval;
	}
}
