/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.create;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.opt4j.core.Genotype;

/**
 * Holds encodings which are generated from the input models. In contrast to {@link Genotype}s, they
 * are not considered as variables by the optimizer.
 * Nevertheless, this provider may also hold {@link Genotype} encodings for the case that an
 * exploration feature is disabled, but it's outputs would be required by dependent features. The
 * provider is responsible to provide the required inputs iff these inputs can be constructed from
 * the input models.
 * 
 * @author diewald
 */
public class InitialEncodingProvider {
	/** Holds the instance. */
	private volatile static InitialEncodingProvider INSTANCE;

	/** Constructor. */
	private InitialEncodingProvider() {
		// Prevent direct instantiation of this singleton.
	}

	/** Returns the instance of this singleton. */
	public static synchronized InitialEncodingProvider getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new InitialEncodingProvider();
		}
		return INSTANCE;
	}

	/**
	 * Maintains a map of {@link IExplorationEncoding}s that can be constructed from the input
	 * models.
	 */
	private Map<Class<? extends IExplorationEncoding>, IExplorationEncoding> inputEncodings =
			new HashMap<>();

	/** Registers a given input encoding that is constructed from the input models. */
	public <T extends IExplorationEncoding> void registerInputEncoding(T encoding) {
		@SuppressWarnings("unchecked") Class<T> encodingType = (Class<T>)encoding.getClass();
		assert (!inputEncodings.containsKey(
				encodingType)) : "You are trying to add an encoding to the input encoding whose type is already registered. This is not supported";
		inputEncodings.put(encodingType, encoding);
	}

	/** Returns the input encoding of the given type. */
	@SuppressWarnings("unchecked")
	public <T extends IExplorationEncoding> T getInputEncoding(Class<T> encodingType) {
		return (T)inputEncodings.get(encodingType);
	}

	/**
	 * Returns the collection of {@link IExplorationEncoding} types constructed from the input
	 * models.
	 */
	public Set<Class<? extends IExplorationEncoding>> getRegisteredEncodingTypes() {
		return inputEncodings.keySet();
	}

	/** Empties the currently held encodings. */
	public void clearEncodings() {
		inputEncodings.clear();
	}
}
