/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;

/**
 * Represents the chosen design of a single safety function, e.g. 1oo2D as defined in the IEC61508
 * standard.
 * <p>
 * Each entry has the following properties:
 * <li>Number of channels.</li>
 * <li>Number of required signals from independent channels to trigger the safety function.</li>
 * <li>Presence of a diagnostic unit connected to the output unit/voter.</li>
 * 
 * @author diewald
 */
public class SafetyFunctionArchEntry<C> implements Cloneable {

	/** References the safety function whose design is represented by {@code this} class. */
	private ISafetyFunctionAdapter<C> safetyFunction;

	/** Defines the number of (independent) channels. */
	private int numChannels = 1;

	/** Defines the number of independent signals to trigger the safety function. */
	private int triggerSignalIdx = 0;

	/**
	 * Number corresponding to the index defining the required number of signal to trigger the
	 * safety function.
	 */
	private int triggerSignalNum = 0;

	/** Presence of a diagnostic unit. */
	// NOTE: Assumption #(diagUnits) == #(channels)!
	private boolean diagnosticUnitPresent = false;

	/** Whether a (present) diagnostic unit is connected to the output "voter". */
	private boolean diagnosticUnitConnected = false;

	/** Constructor. */
	public SafetyFunctionArchEntry(ISafetyFunctionAdapter<C> safetyFunction, int numChannels,
			int numTriggerSignals, boolean diagnosisUnitPresent, boolean diagnosticUnitConnected) {
		this.safetyFunction = safetyFunction;
		this.numChannels = numChannels;
		this.triggerSignalNum = numTriggerSignals;
		for(int i = 0; i < safetyFunction.getAllAllowedTriggerSignalNumbers().size(); i++) {
			if(safetyFunction.getAllowedChannelNumbers()[i] == triggerSignalNum) {
				this.triggerSignalIdx = i;
				break;
			}
		}
		this.diagnosticUnitPresent = diagnosisUnitPresent;
		this.diagnosticUnitConnected = diagnosticUnitConnected;
	}

	/** Returns the safety function whose design is represented by {@code this} class. */
	public ISafetyFunctionAdapter<C> getSafetyFunctionAdapter() {
		return safetyFunction;
	}

	/** Returns the number of channels of {@code this} design of the safety function. */
	public int getNumChannels() {
		return numChannels;
	}

	/** Sets the number of channels that realize {@code this} safety function. */
	// TODO: Guard to check valid numbers.
	public void setNumChannels(int numChannels) {
		this.numChannels = numChannels;

		// Also adjust/align the trigger signals.
		int[] triggerSigNum = safetyFunction.getAllowedTriggerSignalNumbers(numChannels);

		for(int i = 0; i < triggerSigNum.length; i++) {
			if(triggerSigNum[i] == this.triggerSignalNum) {
				this.triggerSignalIdx = i;
				return;
			}
		}

		this.triggerSignalIdx = triggerSigNum.length - 1;
		this.triggerSignalNum = triggerSigNum[triggerSignalIdx];
	}

	/**
	 * Returns the number of signal of {@code this} architecture of the safety function that is
	 * required to trigger the safety function.
	 */
	public int getNumTriggerSignal() {
		int[] triggerSigNum = safetyFunction.getAllowedTriggerSignalNumbers(numChannels);
		return triggerSigNum[triggerSignalIdx];
	}

	/**
	 * Returns the index of the of the current selection of required signals that trigger
	 * {@code this} safety function. The index is defined on the array of trigger signals subject to
	 * the number of channels, i.e. it depends on the selected safety architecture.
	 */
	public int getTriggerSignalIdx() {
		return triggerSignalIdx;
	}

	/**
	 * Sets the index of the array defining the required signals that trigger {@code this} safety
	 * function to the given value. The number of allowed trigger signal is subject ot the selected
	 * number of channels.
	 */
	// TODO: Guard to check valid numbers.
	public void setTriggerSignalIdx(int triggerSignalIdx) {
		int[] triggerSigNum = safetyFunction.getAllowedTriggerSignalNumbers(numChannels);
		if(triggerSignalIdx <= triggerSigNum.length) {
			this.triggerSignalIdx = triggerSignalIdx;
			this.triggerSignalNum =
					safetyFunction.getAllowedTriggerSignalNumbers(numChannels)[triggerSignalIdx];
		}
	}

	/**
	 * Returns whether a diagnosis unit is present in {@code this} design of the safety
	 * function.
	 */
	public boolean isDiagnosicUnitPresent() {
		return diagnosticUnitPresent;
	}

	/** Whether to use a not to use a diagnostic unit. */
	// TODO: Guard to check valid numbers.
	public void setDiagnosicUnitPresent(boolean diagnosicUnitPresent) {
		this.diagnosticUnitPresent = diagnosicUnitPresent;
	}

	/**
	 * Returns whether a diagnosis unit is connected to a voter/output unit in {@code this} design
	 * of the safety function.
	 */
	public boolean isDiagnosicUnitConnected() {
		return diagnosticUnitConnected;
	}

	/** Whether a diagnostic unit shall be connected to the output unit / voter. */
	public void setDiagnosicUnitConnected(boolean diagnosticUnitConnected) {
		this.diagnosticUnitConnected = diagnosticUnitConnected;
	}

	/** {@inheritDoc} */
	@Override
	public SafetyFunctionArchEntry<C> clone() {
		return new SafetyFunctionArchEntry<C>(safetyFunction, numChannels, triggerSignalIdx,
				diagnosticUnitPresent, diagnosticUnitConnected);
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return "SfArch: " + getNumTriggerSignal() + "oo" + getNumChannels() +
				(isDiagnosicUnitPresent() ? (isDiagnosicUnitConnected() ? "D" : "DU") : "") + "\n";
	}
}
