/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalReplicationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalSeparationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.decode.ConstraintGenerationDecoder;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.platform.hierarchic.model.tile.Tile;

import com.google.inject.Inject;

/**
 * Generates allocation and replication constraints for the safety architectures of the safety
 * function of a given {@link SafetyFunctionArchEncoding}.
 * 
 * @author diewald
 */
public class SFMappingConstraintDecoder<C>
		extends ComposableDecoderBase<SafetyFunctionArchEncoding<C>, AbstractTaskMappingEncoding>
		implements ConstraintGenerationDecoder {
	/** Constructs a new decoder that instantiates {@link AbstractTaskMappingEncoding}. */
	@Inject
	public SFMappingConstraintDecoder() {
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding decodeComposable(SafetyFunctionArchEncoding<C> genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding updateComposable(SafetyFunctionArchEncoding<C> inGenotype,
			AbstractTaskMappingEncoding outGenotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) {
		for(SafetyFunctionArchEntry<C> sFAEntry : inGenotype.getAllEntries()) {
			int channelNumber = sFAEntry.getNumChannels();

			Collection<InternalReplicationConstraint> replicationConstraints = new HashSet<>();
			for(ITaskAdapter<?> comp : sFAEntry.getSafetyFunctionAdapter().getAllComponents()) {
				replicationConstraints
						.add(new InternalReplicationConstraint(comp, channelNumber, channelNumber));
			}

			outGenotype.replaceConstraints(InternalReplicationConstraint.class,
					replicationConstraints);
		}

		outGenotype.replaceConstraints(InternalSeparationConstraint.class,
				createSeparationConstraints(inGenotype));

		return outGenotype;
	}

	/**
	 * Creates {@link InternalSeparationConstraint}s for the sets of realizing, diagnostic, and
	 * output units of each safety function.
	 */
	private Collection<InternalSeparationConstraint>
			createSeparationConstraints(SafetyFunctionArchEncoding<C> sfEnc) {
		Collection<InternalSeparationConstraint> sepConstraints = new HashSet<>();

		for(SafetyFunctionArchEntry<C> sfEntry : sfEnc.getAllEntries()) {
			ISafetyFunctionAdapter<C> sfAdp = sfEntry.getSafetyFunctionAdapter();

			// TODO: We should not be aware of the Tile type here.
			// FIXME: We need a "better" machanism than the fault containment region. Also,
			sepConstraints.add(
					new InternalSeparationConstraint(sfAdp.getRealizingComponents(), Tile.class));
			sepConstraints.add(
					new InternalSeparationConstraint(sfAdp.getDiagnosticComponents(), Tile.class));
			sepConstraints.add(new InternalSeparationConstraint(
					Arrays.asList(sfAdp.getOutputComponent()), Tile.class));
		}

		return sepConstraints;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> requiredTypes = new HashSet<>();
		requiredTypes.add(SafetyFunctionArchEncoding.class);
		return requiredTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return SafetyFunctionArchEncoding.class;
	}
}
