/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.disableNotificationsOf;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.enableNotificationsOf;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementReference;

/**
 * Handles operations on EMF models that are problematic when used from different threads. In order
 * to provide the required synchronization before the model is accessed, the
 * {@link EMFModelConcurrencyHandler} is implemented as a Singleton.
 * 
 * @author diewald
 */
public class EMFModelConcurrencyHandler {

	/** Holds the instance of the model handler. */
	private volatile static EMFModelConcurrencyHandler INSTANCE;

	/**
	 * Bookkeeping of objects (value) which have requested to suppress notifications of a model
	 * element (key, EObject).
	 */
	private Set<EObject> suppressedNotificationModelEditing;

	/** Constructor. */
	private EMFModelConcurrencyHandler() {
		// Prevent direct instantiation.
		suppressedNotificationModelEditing = new HashSet<>();
	}

	/** The instance (singleton) of the EvaluatorWrapper factory. */
	public static synchronized EMFModelConcurrencyHandler getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new EMFModelConcurrencyHandler();
		}
		return INSTANCE;
	}

	/**
	 * Helper to reset reference of a temporary object (that is not hooked to the {@link Resource}
	 * without triggering a model modification.
	 */
	public synchronized void resetReference(IModelElementReference ref) {
		if(ref == null) {
			return;
		}

		IModelElement refElement = ref.getReference();
		// ITopLevelElement modelContext =
		// IPersistencyService.getInstance().getTopLevelElementFor(refElement);
		// if(modelContext == null) {
		disableNotificationsOf(refElement);
		ref.setReference(null);
		enableNotificationsOf(refElement);
		// } else {
		// List<IModelElementReference> refByList = refElement.getReferencedBy();
		// modelContext.runAsNonDirtyingCommand(() -> refByList.remove(ref));
		// // modelContext.removeLastCommandFromStack();
		// }
	}

	/**
	 * Helper to set a {@link IModelElementReference} from a temporary and programmatically
	 * generated model without triggering a model modification.
	 * If this method is not used, a notification is triggered from the referenced model because a
	 * back reference is set (Bi-Reference) in the referenced model.
	 */
	public synchronized void setReference(IModelElementReference ref,
			IModelElement referencedElement) {
		if(ref == null) {
			return;
		}
		//
		// ITopLevelElement modelContext =
		// IPersistencyService.getInstance().getTopLevelElementFor(referencedElement);
		// if(modelContext == null) {
		disableNotificationsOf(referencedElement);
		ref.setReference(referencedElement);
		enableNotificationsOf(referencedElement);
		// } else {
		// modelContext.runAsNonDirtyingCommand(() -> ref.setReference(referencedElement));
		// // modelContext.removeLastCommandFromStack();
		// }
	}

	/**
	 * Determines whether there are still objects that request the given model not to emit
	 * notifications.
	 */
	public synchronized boolean isModelEdited(EObject model) {
		return suppressedNotificationModelEditing.contains(model);
	}

	/**
	 * Deletes the suppress notification request of the given {@code editingObj} and enables the
	 * notifications of the given model if there are no more requesters to suppress notifications.
	 */
	public synchronized void tryEnableNotificationsSyncOf(EObject model) {
		getInstance().removeSuppressedNotificationModelEditing(model);
		// if(!getInstance().isModelEdited(model)) {
		enableNotificationsOf(model);
		// }
	}

	/**
	 * Adds a suppress notification request of the given {@code editingObj} and disables the
	 * notifications of the given model if needed.
	 */
	public synchronized void tryDisableNotificationsSyncOf(EObject model) {
		getInstance().addSuppressedNotificationModelEditing(model);
		// if(!getInstance().isModelEdited(model)) {
		disableNotificationsOf(model);
		// }
	}

	/**
	 * Adds the given {@code editingObj} as a requester to suppress notifications from the given
	 * {@code model}.
	 */
	private synchronized void addSuppressedNotificationModelEditing(EObject model) {
		suppressedNotificationModelEditing.add(model);
	}

	/**
	 * Removes the given {@code editingObj} as a requester to suppress notifications from the given
	 * {@code model}.
	 */
	private synchronized void removeSuppressedNotificationModelEditing(EObject model) {
		suppressedNotificationModelEditing.remove(model);
	}
}
