/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.comm;

import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.constraint.InternalIsolatedCommunicationSet;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;

/**
 * Responsible for generating the {@link Message}s that must be sent between different
 * {@link IMappingEntry}s to satisfy the data dependencies that are encoded in the input logical
 * architecture.
 * 
 * @author diewald
 */
public class MessageGenerator<S extends TaskMappingEntry<ITaskAdapter<?>, Partition>, T extends TaskMappingEncoding<ITaskAdapter<?>, S>> {
	/**
	 * Generates messages from the given {#link encoding}. This methods evaluates the {#link
	 * encoding}
	 */
	public Collection<Message> generateMessages(T encoding,
			AbstractTaskGraphEncoding<AbstractTaskMappingEncoding, ?> atgEnc) {
		Collection<Message> generatedMessages = new HashSet<Message>();

		// TODO: represent the signals in the task graph: do not use the systemModelAdapter here.
		// Consequence: there may be multiple edges between two vertices (tasks).
		for(S entry : encoding.getMappingEntries()) {
			ITaskAdapter<?> senderComponent = entry.getSource();
			if(senderComponent.isInstantiation()) {
				senderComponent = senderComponent.getReplacedComponent();
			}

			TaskGraph tg = atgEnc.getTaskGraphOf(senderComponent);

			@SuppressWarnings("unchecked") Collection<ISignalAdapter<?, ?>> emittedMessages =
					new ArrayList<>();
			tg.getGraph().outgoingEdgesOf(senderComponent)
					.forEach(e -> emittedMessages.add(e.getSignalAdapter()));

			Collection<InternalIsolatedCommunicationSet> allIsoCommConstr =
					encoding.getConstraintsOf(InternalIsolatedCommunicationSet.class);
			Collection<InternalIsolatedCommunicationSet> otherIsoCommConstr =
					getOtherIsolationCommConstraints(encoding, entry);

			for(ISignalAdapter<?, ?> currentMessageAdapter : emittedMessages) {
				// Create a Collection of dependent MappingEntries, for which messages are generated
				// (if required).
				// TODO: Get rid of the AbstractTaskMappingEntry(s) and use S only. Note that guava
				// generic type reflection will be required!
				Collection<S> mappingEntriesOfSuccesors =
						getSuccessorMappingEntries(encoding, currentMessageAdapter);
				Message generatedMessage = generateMessageForSenderEntry(currentMessageAdapter,
						(InstantiatedTaskMappingEntry)entry, mappingEntriesOfSuccesors,
						allIsoCommConstr, otherIsoCommConstr);
				if(generatedMessage != null) {
					generatedMessages.add(generatedMessage);
				}
			}
		}

		return generatedMessages;
	}

	/**
	 * Creates a {@link Collection} which contains all {@link IMappingEntry}s that are causally
	 * depend on the given sender {@link ITaskAdapter}.
	 */
	private Collection<S> getSuccessorMappingEntries(T encoding,
			ISignalAdapter<?, ?> emittedMessage) {
		Collection<S> mappingEntriesOfSuccesors = new HashSet<S>();
		Collection<ITaskAdapter<?>> realSuccessorComponents = new HashSet<>();

		for(ITaskAdapter<?> successorComponent : emittedMessage.getTargets()) {
			if(successorComponent.isInstantiationRequired()) {
				// Since the generated message refer to the components in the system model adapter,
				// we must search the encoding for equivalent components in this special case
				// (instantiation).
				realSuccessorComponents.addAll(successorComponent.getReplacementComponents());
			} else {
				realSuccessorComponents.add(successorComponent);
			}
		}

		for(ITaskAdapter<?> successorComponent : realSuccessorComponents) {
			Collection<ITaskAdapter<?>> encComps = filterSet(encoding.getRequesters(),
					c -> c.getObject() == successorComponent.getObject());
			for(ITaskAdapter<?> encComp : encComps) {
				mappingEntriesOfSuccesors.addAll(encoding.getMappingEntriesOf(encComp));
			}
		}
		return mappingEntriesOfSuccesors;
	}

	/**
	 * Creates a new Message for the given {#link senderMappingEntry} and its receiver
	 * {@link IMappingEntry}s if the sender and the receiver are not mapped onto the same
	 * {@link IExecutionUnitAdapter}.
	 */
	@SuppressWarnings("unchecked")
	private Message generateMessageForSenderEntry(ISignalAdapter<?, ?> senderChannel,
			InstantiatedTaskMappingEntry senderMappingEntry,
			Collection<S> mappingEntriesOfSuccesors,
			Collection<InternalIsolatedCommunicationSet> allConstraints,
			Collection<InternalIsolatedCommunicationSet> otherConstraints) {

		Collection<S> receiverEntries = new HashSet<S>();
		for(S mappingEntryReceiver : mappingEntriesOfSuccesors) {
			// TODO: knowledge about prev steps.
			AbstractTaskMappingEntry constrEntry =
					mappingEntryReceiver.getAssociatedElement(AbstractTaskMappingEntry.class);
			if((senderMappingEntry.getTarget() != mappingEntryReceiver.getTarget())) {
				if(isPartOfIsolationCommConstraint(allConstraints, senderMappingEntry) &&
						!otherConstraints.stream().anyMatch(c -> c.contains(constrEntry))) {
					receiverEntries.add(mappingEntryReceiver);
				} else if(!isPartOfIsolationCommConstraint(allConstraints, senderMappingEntry)) {
					receiverEntries.add(mappingEntryReceiver);
				}
			}
		}

		if(!receiverEntries.isEmpty()) {
			return new Message(senderChannel, senderMappingEntry,
					(Collection<? extends InstantiatedTaskMappingEntry>)receiverEntries);
		}
		return null;
	}

	/**
	 * Returns whether the given {@link IMappingEntry} is included in any of the given
	 * {@link InternalIsolatedCommunicationSet}s.
	 */
	private boolean isPartOfIsolationCommConstraint(
			Collection<InternalIsolatedCommunicationSet> isoComm,
			InstantiatedTaskMappingEntry entry) {
		AbstractTaskMappingEntry actualEntry =
				entry.getAssociatedElement(AbstractTaskMappingEntry.class);

		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(constr.contains(actualEntry)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Returns a collection of all {@link InternalIsolatedCommunicationSet} constraints that do not
	 * include the given {@link IMappingEntry}.
	 */
	private Collection<InternalIsolatedCommunicationSet>
			getOtherIsolationCommConstraints(T outGenotype, S entry) {
		// TODO: here we use the knowledge that we have instantiated task mapping entries.
		IMappingEntry<?, ?> actualEntry =
				entry.getAssociatedElement(AbstractTaskMappingEntry.class);

		Collection<InternalIsolatedCommunicationSet> otherConstraints = new HashSet<>();
		Collection<InternalIsolatedCommunicationSet> isoComm =
				outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class);
		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(!constr.getIsolatedCommunicationSet().contains(actualEntry)) {
				otherConstraints.add(constr);
			}
		}

		return otherConstraints;
	}
}
