/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.guava;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.performNearestClassLookup;

import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multiset;

/**
 * {@link Multimap} whose keys are types. Further, this map performs a lookup in the type hierarchy
 * if it does not contain the given type (key). The lookup is performed for {@code get} and
 * {@code put} methods.
 * 
 * @author diewald
 */
public class SubClassToInstanceMultiMap<K, V> implements Multimap<Class<? extends K>, V> {

	/** Backend {@link Multimap} that holds the actual keys and values. */
	Multimap<Class<? extends K>, V> delegateMap = HashMultimap.create();

	/** Bound up to which a lookup is performed. */
	final Class<? extends K> typeBound;

	/**
	 * Constructor. The type parameter defines how far (in the inheritance tree) a lookup of
	 * alternative keys shall be performed if {@code this} {@link Multimap} does not directly
	 * contain the given key. The given type bound is considered to be exclusive, i.e. if the lookup
	 * returns the same type as the bound, its result will be ignored.
	 * 
	 * @param typeBound
	 *            Type bound up to which a lookup for matching keys is performed if the backend map
	 *            does not contain a given key directly.
	 */
	public SubClassToInstanceMultiMap(Class<? extends K> typeBound) {
		super();
		this.typeBound = typeBound;
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return delegateMap.size();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEmpty() {
		return delegateMap.isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsKey(Object key) {
		return delegateMap.containsKey(key);
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsValue(Object value) {
		return delegateMap.containsValue(value);
	}

	/** {@inheritDoc} */
	@Override
	public boolean containsEntry(Object key, Object value) {
		return delegateMap.containsEntry(key, value);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public boolean put(Class<? extends K> key, V value) {
		Class<?> existingKey = performNearestClassLookup(key, delegateMap.keySet());
		if(typeBound.isAssignableFrom(existingKey)) {
			key = (Class<? extends K>)existingKey;
		}
		return delegateMap.put(key, value);
	}

	/** {@inheritDoc} */
	@Override
	public boolean remove(Object key, Object value) {
		return delegateMap.remove(key, value);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public boolean putAll(Class<? extends K> key, Iterable<? extends V> values) {
		Class<?> existingKey = performNearestClassLookup(key, delegateMap.keySet());
		if(typeBound.isAssignableFrom(existingKey)) {
			key = (Class<? extends K>)existingKey;
		}
		return delegateMap.putAll(key, values);
	}

	/** {@inheritDoc} */
	@Override
	public boolean putAll(Multimap<? extends Class<? extends K>, ? extends V> multimap) {
		return delegateMap.putAll(multimap);
	}

	/** {@inheritDoc} */
	@Override
	public Collection<V> replaceValues(Class<? extends K> key, Iterable<? extends V> values) {

		return delegateMap.replaceValues(key, values);
	}

	/** {@inheritDoc} */
	@Override
	public Collection<V> removeAll(Object key) {
		return delegateMap.removeAll(key);
	}

	/** {@inheritDoc} */
	@Override
	public void clear() {
		delegateMap.clear();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<V> get(Class<? extends K> key) {
		Collection<V> retVal = delegateMap.get(key);
		if(retVal.isEmpty()) {
			retVal = ExplorationAlgUtils.performNearestClassLookup(key, delegateMap);
		}
		return retVal;
	}

	/** {@inheritDoc} */
	@Override
	public Set<Class<? extends K>> keySet() {
		return delegateMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Multiset<Class<? extends K>> keys() {
		return delegateMap.keys();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<V> values() {
		return delegateMap.values();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Entry<Class<? extends K>, V>> entries() {
		return delegateMap.entries();
	}

	/** {@inheritDoc} */
	@Override
	public Map<Class<? extends K>, Collection<V>> asMap() {
		return delegateMap.asMap();
	}
}
